@@ -1,4 +1,4 @@
-# Copyright (c) 2001-2008 Twisted Matrix Laboratories.
+# Copyright (c) Twisted Matrix Laboratories.
 # See LICENSE for details.
 
 """
@@ -7,12 +7,12 @@ Test cases for twisted.mail.smtp module.
 
 from zope.interface import implements
 
+from twisted.python.util import LineLog
 from twisted.trial import unittest, util
 from twisted.protocols import basic, loopback
 from twisted.mail import smtp
 from twisted.internet import defer, protocol, reactor, interfaces
 from twisted.internet import address, error, task
-from twisted.test.test_protocols import StringIOWithoutClosing
 from twisted.test.proto_helpers import StringTransport
 
 from twisted import cred
@@ -41,21 +41,57 @@ try:
 except ImportError:
     from StringIO import StringIO
 
+
 def spameater(*spam, **eggs):
     return None
 
-class DummyMessage:
 
+
+class BrokenMessage(object):
+    """
+    L{BrokenMessage} is an L{IMessage} which raises an unexpected exception
+    from its C{eomReceived} method.  This is useful for creating a server which
+    can be used to test client retry behavior.
+    """
+    implements(smtp.IMessage)
+
+    def __init__(self, user):
+        pass
+
+
+    def lineReceived(self, line):
+        pass
+
+
+    def eomReceived(self):
+        raise RuntimeError("Some problem, delivery is failing.")
+
+
+    def connectionLost(self):
+        pass
+
+
+
+class DummyMessage(object):
+    """
+    L{BrokenMessage} is an L{IMessage} which saves the message delivered to it
+    to its domain object.
+
+    @ivar domain: A L{DummyDomain} which will be used to store the message once
+        it is received.
+    """
     def __init__(self, domain, user):
         self.domain = domain
         self.user = user
         self.buffer = []
 
+
     def lineReceived(self, line):
         # Throw away the generated Received: header
         if not re.match('Received: From yyy.com \(\[.*\]\) by localhost;', line):
             self.buffer.append(line)
 
+
     def eomReceived(self):
         message = '\n'.join(self.buffer) + '\n'
         self.domain.messages[self.user.dest.local].append(message)
@@ -64,20 +100,28 @@ class DummyMessage:
         return deferred
 
 
-class DummyDomain:
 
-   def __init__(self, names):
-       self.messages = {}
-       for name in names:
-           self.messages[name] = []
+class DummyDomain(object):
+    """
+    L{DummyDomain} is an L{IDomain} which keeps track of messages delivered to
+    it in memory.
+    """
+    def __init__(self, names):
+        self.messages = {}
+        for name in names:
+            self.messages[name] = []
+
+
+    def exists(self, user):
+        if user.dest.local in self.messages:
+            return defer.succeed(lambda: self.startMessage(user))
+        return defer.fail(smtp.SMTPBadRcpt(user))
+
+
+    def startMessage(self, user):
+        return DummyMessage(self, user)
 
-   def exists(self, user):
-       if self.messages.has_key(user.dest.local):
-           return defer.succeed(lambda: self.startMessage(user))
-       return defer.fail(smtp.SMTPBadRcpt(user))
 
-   def startMessage(self, user):
-       return DummyMessage(self, user)
 
 class SMTPTestCase(unittest.TestCase):
 
@@ -90,11 +134,15 @@ Someone set up us the bomb!\015
     mbox = {'foo': ['Subject: urgent\n\nSomeone set up us the bomb!\n']}
 
     def setUp(self):
+        """
+        Create an in-memory mail domain to which messages may be delivered by
+        tests and create a factory and transport to do the delivering.
+        """
         self.factory = smtp.SMTPFactory()
         self.factory.domains = {}
         self.factory.domains['baz.com'] = DummyDomain(['foo'])
-        self.output = StringIOWithoutClosing()
-        self.transport = protocol.FileWrapper(self.output)
+        self.transport = StringTransport()
+
 
     def testMessages(self):
         from twisted.mail import protocols
@@ -125,25 +173,43 @@ Goodbye
 '''
 
 class MyClient:
-    def __init__(self):
-        self.mail = 'moshez@foo.bar', ['moshez@foo.bar'], mail
+    def __init__(self, messageInfo=None):
+        if messageInfo is None:
+            messageInfo = (
+                'moshez@foo.bar', ['moshez@foo.bar'], StringIO(mail))
+        self._sender = messageInfo[0]
+        self._recipient = messageInfo[1]
+        self._data = messageInfo[2]
+
 
     def getMailFrom(self):
-        return self.mail[0]
+        return self._sender
+
 
     def getMailTo(self):
-        return self.mail[1]
+        return self._recipient
+
 
     def getMailData(self):
-        return StringIO(self.mail[2])
+        return self._data
+
+
+    def sendError(self, exc):
+        self._error = exc
+
 
     def sentMail(self, code, resp, numOk, addresses, log):
-        self.mail = None, None, None
+        # Prevent another mail from being sent.
+        self._sender = None
+        self._recipient = None
+        self._data = None
+
+
 
 class MySMTPClient(MyClient, smtp.SMTPClient):
-    def __init__(self):
+    def __init__(self, messageInfo=None):
         smtp.SMTPClient.__init__(self, 'foo.baz')
-        MyClient.__init__(self)
+        MyClient.__init__(self, messageInfo)
 
 class MyESMTPClient(MyClient, smtp.ESMTPClient):
     def __init__(self, secret = '', contextFactory = None):
@@ -202,6 +268,29 @@ class FakeSMTPServer(basic.LineReceiver):
 
 
 class SMTPClientTestCase(unittest.TestCase, LoopbackMixin):
+    """
+    Tests for L{smtp.SMTPClient}.
+    """
+
+    def test_timeoutConnection(self):
+        """
+        L{smtp.SMTPClient.timeoutConnection} calls the C{sendError} hook with a
+        fatal L{SMTPTimeoutError} with the current line log.
+        """
+        error = []
+        client = MySMTPClient()
+        client.sendError = error.append
+        client.makeConnection(StringTransport())
+        client.lineReceived("220 hello")
+        client.timeoutConnection()
+        self.assertIsInstance(error[0], smtp.SMTPTimeoutError)
+        self.assertTrue(error[0].isFatal)
+        self.assertEqual(
+            str(error[0]),
+            "Timeout waiting for SMTP server response\n"
+            "<<< 220 hello\n"
+            ">>> HELO foo.baz\n")
+
 
     expected_output = [
         'HELO foo.baz', 'MAIL FROM:<moshez@foo.bar>',
@@ -209,8 +298,12 @@ class SMTPClientTestCase(unittest.TestCase, LoopbackMixin):
         'Subject: hello', '', 'Goodbye', '.', 'RSET'
     ]
 
-    def testMessages(self):
-        # this test is disabled temporarily
+    def test_messages(self):
+        """
+        L{smtp.SMTPClient} sends I{HELO}, I{MAIL FROM}, I{RCPT TO}, and I{DATA}
+        commands based on the return values of its C{getMailFrom},
+        C{getMailTo}, and C{getMailData} methods.
+        """
         client = MySMTPClient()
         server = FakeSMTPServer()
         d = self.loopback(server, client)
@@ -218,6 +311,78 @@ class SMTPClientTestCase(unittest.TestCase, LoopbackMixin):
                       self.assertEquals(server.buffer, self.expected_output))
         return d
 
+
+    def test_transferError(self):
+        """
+        If there is an error while producing the message body to the
+        connection, the C{sendError} callback is invoked.
+        """
+        client = MySMTPClient(
+            ('alice@example.com', ['bob@example.com'], StringIO("foo")))
+        transport = StringTransport()
+        client.makeConnection(transport)
+        client.dataReceived(
+            '220 Ok\r\n' # Greeting
+            '250 Ok\r\n' # EHLO response
+            '250 Ok\r\n' # MAIL FROM response
+            '250 Ok\r\n' # RCPT TO response
+            '354 Ok\r\n' # DATA response
+            )
+
+        # Sanity check - a pull producer should be registered now.
+        self.assertNotIdentical(transport.producer, None)
+        self.assertFalse(transport.streaming)
+
+        # Now stop the producer prematurely, meaning the message was not sent.
+        transport.producer.stopProducing()
+
+        # The sendError hook should have been invoked as a result.
+        self.assertIsInstance(client._error, Exception)
+
+
+    def test_sendFatalError(self):
+        """
+        If L{smtp.SMTPClient.sendError} is called with an L{SMTPClientError}
+        which is fatal, it disconnects its transport without writing anything
+        more to it.
+        """
+        client = smtp.SMTPClient(None)
+        transport = StringTransport()
+        client.makeConnection(transport)
+        client.sendError(smtp.SMTPClientError(123, "foo", isFatal=True))
+        self.assertEqual(transport.value(), "")
+        self.assertTrue(transport.disconnecting)
+
+
+    def test_sendNonFatalError(self):
+        """
+        If L{smtp.SMTPClient.sendError} is called with an L{SMTPClientError}
+        which is not fatal, it sends C{"QUIT"} and waits for the server to
+        close the connection.
+        """
+        client = smtp.SMTPClient(None)
+        transport = StringTransport()
+        client.makeConnection(transport)
+        client.sendError(smtp.SMTPClientError(123, "foo", isFatal=False))
+        self.assertEqual(transport.value(), "QUIT\r\n")
+        self.assertFalse(transport.disconnecting)
+
+
+    def test_sendOtherError(self):
+        """
+        If L{smtp.SMTPClient.sendError} is called with an exception which is
+        not an L{SMTPClientError}, it disconnects its transport without
+        writing anything more to it.
+        """
+        client = smtp.SMTPClient(None)
+        transport = StringTransport()
+        client.makeConnection(transport)
+        client.sendError(Exception("foo"))
+        self.assertEqual(transport.value(), "")
+        self.assertTrue(transport.disconnecting)
+
+
+
 class DummySMTPMessage:
 
     def __init__(self, protocol, users):
@@ -237,6 +402,8 @@ class DummySMTPMessage:
         self.protocol.message[tuple(recipients)] = (helo, origin, recipients, message)
         return defer.succeed("saved")
 
+
+
 class DummyProto:
     def connectionMade(self):
         self.dummyMixinBase.connectionMade(self)
@@ -249,12 +416,14 @@ class DummyProto:
         return None
 
     def validateTo(self, user):
-        self.delivery = DummyDelivery()
+        self.delivery = SimpleDelivery(None)
         return lambda: self.startMessage([user])
 
     def validateFrom(self, helo, origin):
         return origin
 
+
+
 class DummySMTP(DummyProto, smtp.SMTP):
     dummyMixinBase = smtp.SMTP
 
@@ -308,19 +477,25 @@ To: foo
                            (helo_, realfrom, realto, msg))))
 
 
-    def testBuffer(self):
-        output = StringIOWithoutClosing()
+    def test_buffer(self):
+        """
+        Exercise a lot of the SMTP client code.  This is a "shotgun" style unit
+        test.  It does a lot of things and hopes that something will go really
+        wrong if it is going to go wrong.  This test should be replaced with a
+        suite of nicer tests.
+        """
+        transport = StringTransport()
         a = self.serverClass()
         class fooFactory:
             domain = 'foo.com'
 
         a.factory = fooFactory()
-        a.makeConnection(protocol.FileWrapper(output))
+        a.makeConnection(transport)
         for (send, expect, msg, msgexpect) in self.data:
             if send:
                 a.dataReceived(send)
-            data = output.getvalue()
-            output.truncate(0)
+            data = transport.value()
+            transport.clear()
             if not re.match(expect, data):
                 raise AssertionError, (send, expect, data)
             if data[:3] == '354':
@@ -331,8 +506,8 @@ To: foo
                 a.dataReceived('.\r\n')
                 # Special case for DATA. Now we want a 250, and then
                 # we compare the messages
-                data = output.getvalue()
-                output.truncate()
+                data = transport.value()
+                transport.clear()
                 resp, msgdata = msgexpect
                 if not re.match(resp, data):
                     raise AssertionError, (resp, data)
@@ -363,7 +538,8 @@ class DummyChecker:
         'testuser': 'testpassword'
     }
 
-    credentialInterfaces = (cred.credentials.IUsernameHashedPassword,)
+    credentialInterfaces = (cred.credentials.IUsernamePassword,
+                            cred.credentials.IUsernameHashedPassword)
 
     def requestAvatarId(self, credentials):
         return defer.maybeDeferred(
@@ -375,24 +551,45 @@ class DummyChecker:
             return username
         raise cred.error.UnauthorizedLogin()
 
-class DummyDelivery:
+
+
+class SimpleDelivery(object):
+    """
+    L{SimpleDelivery} is a message delivery factory with no interesting
+    behavior.
+    """
     implements(smtp.IMessageDelivery)
 
-    def validateTo(self, user):
-        return user
+    def __init__(self, messageFactory):
+        self._messageFactory = messageFactory
+
+
+    def receivedHeader(self, helo, origin, recipients):
+        return None
+
 
     def validateFrom(self, helo, origin):
         return origin
 
-    def receivedHeader(*args):
-        return None
+
+    def validateTo(self, user):
+        return lambda: self._messageFactory(user)
+
+
 
 class DummyRealm:
     def requestAvatar(self, avatarId, mind, *interfaces):
-        return smtp.IMessageDelivery, DummyDelivery(), lambda: None
+        return smtp.IMessageDelivery, SimpleDelivery(None), lambda: None
+
+
 
 class AuthTestCase(unittest.TestCase, LoopbackMixin):
-    def testAuth(self):
+    def test_crammd5Auth(self):
+        """
+        L{ESMTPClient} can authenticate using the I{CRAM-MD5} SASL mechanism.
+
+        @see: U{http://tools.ietf.org/html/rfc2195}
+        """
         realm = DummyRealm()
         p = cred.portal.Portal(realm)
         p.registerChecker(DummyChecker())
@@ -401,13 +598,65 @@ class AuthTestCase(unittest.TestCase, LoopbackMixin):
         server.portal = p
         client = MyESMTPClient('testpassword')
 
-        cAuth = imap4.CramMD5ClientAuthenticator('testuser')
+        cAuth = smtp.CramMD5ClientAuthenticator('testuser')
         client.registerAuthenticator(cAuth)
 
         d = self.loopback(server, client)
         d.addCallback(lambda x : self.assertEquals(server.authenticated, 1))
         return d
 
+
+    def test_loginAuth(self):
+        """
+        L{ESMTPClient} can authenticate using the I{LOGIN} SASL mechanism.
+
+        @see: U{http://sepp.oetiker.ch/sasl-2.1.19-ds/draft-murchison-sasl-login-00.txt}
+        """
+        realm = DummyRealm()
+        p = cred.portal.Portal(realm)
+        p.registerChecker(DummyChecker())
+
+        server = DummyESMTP({'LOGIN': imap4.LOGINCredentials})
+        server.portal = p
+        client = MyESMTPClient('testpassword')
+
+        cAuth = smtp.LOGINAuthenticator('testuser')
+        client.registerAuthenticator(cAuth)
+
+        d = self.loopback(server, client)
+        d.addCallback(lambda x: self.assertTrue(server.authenticated))
+        return d
+
+
+    def test_loginAgainstWeirdServer(self):
+        """
+        When communicating with a server which implements the I{LOGIN} SASL
+        mechanism using C{"Username:"} as the challenge (rather than C{"User
+        Name\\0"}), L{ESMTPClient} can still authenticate successfully using
+        the I{LOGIN} mechanism.
+        """
+        realm = DummyRealm()
+        p = cred.portal.Portal(realm)
+        p.registerChecker(DummyChecker())
+
+        class WeirdLOGIN(imap4.LOGINCredentials):
+            def __init__(self):
+                imap4.LOGINCredentials.__init__(self)
+                self.challenges[1] = 'Username:'
+
+        server = DummyESMTP({'LOGIN': WeirdLOGIN})
+        server.portal = p
+
+        client = MyESMTPClient('testpassword')
+        cAuth = smtp.LOGINAuthenticator('testuser')
+        client.registerAuthenticator(cAuth)
+
+        d = self.loopback(server, client)
+        d.addCallback(lambda x: self.assertTrue(server.authenticated))
+        return d
+
+
+
 class SMTPHelperTestCase(unittest.TestCase):
     def testMessageID(self):
         d = {}
@@ -468,11 +717,6 @@ class SMTPHelperTestCase(unittest.TestCase):
         Similar to L{test_encodeWithErrors}, but for C{str.decode}.
         """
         bytes = 'Hello world'
-        self.assertEqual(
-            smtp._slowXTextDecode(bytes, 'strict'),
-            (bytes.decode('xtext'), len(bytes)))
-        # This might be the same as _slowXTextDecode, but it might also be the
-        # fast version instead.
         self.assertEqual(
             smtp.xtext_decode(bytes, 'strict'),
             (bytes.decode('xtext'), len(bytes)))
@@ -512,15 +756,18 @@ if not interfaces.IReactorSSL.providedBy(reactor):
         case.skip = "Reactor doesn't support SSL"
 
 class EmptyLineTestCase(unittest.TestCase):
-    def testEmptyLineSyntaxError(self):
+    def test_emptyLineSyntaxError(self):
+        """
+        If L{smtp.SMTP} receives an empty line, it responds with a 500 error
+        response code and a message about a syntax error.
+        """
         proto = smtp.SMTP()
-        output = StringIOWithoutClosing()
-        transport = protocol.FileWrapper(output)
+        transport = StringTransport()
         proto.makeConnection(transport)
         proto.lineReceived('')
         proto.setTimeout(None)
 
-        out = output.getvalue().splitlines()
+        out = transport.value().splitlines()
         self.assertEquals(len(out), 2)
         self.failUnless(out[0].startswith('220'))
         self.assertEquals(out[1], "500 Error: bad syntax")
@@ -581,6 +828,165 @@ class TimeoutTestCase(unittest.TestCase, LoopbackMixin):
         return self._timeoutTest(onDone, clientFactory)
 
 
+    def test_resetTimeoutWhileSending(self):
+        """
+        The timeout is not allowed to expire after the server has accepted a
+        DATA command and the client is actively sending data to it.
+        """
+        class SlowFile:
+            """
+            A file-like which returns one byte from each read call until the
+            specified number of bytes have been returned.
+            """
+            def __init__(self, size):
+                self._size = size
+
+            def read(self, max=None):
+                if self._size:
+                    self._size -= 1
+                    return 'x'
+                return ''
+
+        failed = []
+        onDone = defer.Deferred()
+        onDone.addErrback(failed.append)
+        clientFactory = smtp.SMTPSenderFactory(
+            'source@address', 'recipient@address',
+            SlowFile(1), onDone, retries=0, timeout=3)
+        clientFactory.domain = "example.org"
+        clock = task.Clock()
+        client = clientFactory.buildProtocol(
+            address.IPv4Address('TCP', 'example.net', 25))
+        client.callLater = clock.callLater
+        transport = StringTransport()
+        client.makeConnection(transport)
+
+        client.dataReceived(
+            "220 Ok\r\n" # Greet the client
+            "250 Ok\r\n" # Respond to HELO
+            "250 Ok\r\n" # Respond to MAIL FROM
+            "250 Ok\r\n" # Respond to RCPT TO
+            "354 Ok\r\n" # Respond to DATA
+            )
+
+        # Now the client is producing data to the server.  Any time
+        # resumeProducing is called on the producer, the timeout should be
+        # extended.  First, a sanity check.  This test is only written to
+        # handle pull producers.
+        self.assertNotIdentical(transport.producer, None)
+        self.assertFalse(transport.streaming)
+
+        # Now, allow 2 seconds (1 less than the timeout of 3 seconds) to
+        # elapse.
+        clock.advance(2)
+
+        # The timeout has not expired, so the failure should not have happened.
+        self.assertEqual(failed, [])
+
+        # Let some bytes be produced, extending the timeout.  Then advance the
+        # clock some more and verify that the timeout still hasn't happened.
+        transport.producer.resumeProducing()
+        clock.advance(2)
+        self.assertEqual(failed, [])
+
+        # The file has been completely produced - the next resume producing
+        # finishes the upload, successfully.
+        transport.producer.resumeProducing()
+        client.dataReceived("250 Ok\r\n")
+        self.assertEqual(failed, [])
+
+        # Verify that the client actually did send the things expected.
+        self.assertEqual(
+            transport.value(),
+            "HELO example.org\r\n"
+            "MAIL FROM:<source@address>\r\n"
+            "RCPT TO:<recipient@address>\r\n"
+            "DATA\r\n"
+            "x\r\n"
+            ".\r\n"
+            # This RSET is just an implementation detail.  It's nice, but this
+            # test doesn't really care about it.
+            "RSET\r\n")
+
+
+
+class MultipleDeliveryFactorySMTPServerFactory(protocol.ServerFactory):
+    """
+    L{MultipleDeliveryFactorySMTPServerFactory} creates SMTP server protocol
+    instances with message delivery factory objects supplied to it.  Each
+    factory is used for one connection and then discarded.  Factories are used
+    in the order they are supplied.
+    """
+    def __init__(self, messageFactories):
+        self._messageFactories = messageFactories
+
+
+    def buildProtocol(self, addr):
+        p = protocol.ServerFactory.buildProtocol(self, addr)
+        p.delivery = SimpleDelivery(self._messageFactories.pop(0))
+        return p
+
+
+
+class SMTPSenderFactoryRetryTestCase(unittest.TestCase):
+    """
+    Tests for the retry behavior of L{smtp.SMTPSenderFactory}.
+    """
+    def test_retryAfterDisconnect(self):
+        """
+        If the protocol created by L{SMTPSenderFactory} loses its connection
+        before receiving confirmation of message delivery, it reconnects and
+        tries to deliver the message again.
+        """
+        recipient = 'alice'
+        message = "some message text"
+        domain = DummyDomain([recipient])
+
+        class CleanSMTP(smtp.SMTP):
+            """
+            An SMTP subclass which ensures that its transport will be
+            disconnected before the test ends.
+            """
+            def makeConnection(innerSelf, transport):
+                self.addCleanup(transport.loseConnection)
+                smtp.SMTP.makeConnection(innerSelf, transport)
+
+        # Create a server which will fail the first message deliver attempt to
+        # it with a 500 and a disconnect, but which will accept a message
+        # delivered over the 2nd connection to it.
+        serverFactory = MultipleDeliveryFactorySMTPServerFactory([
+                BrokenMessage,
+                lambda user: DummyMessage(domain, user)])
+        serverFactory.protocol = CleanSMTP
+        serverPort = reactor.listenTCP(0, serverFactory, interface='127.0.0.1')
+        serverHost = serverPort.getHost()
+        self.addCleanup(serverPort.stopListening)
+
+        # Set up a client to try to deliver a message to the above created
+        # server.
+        sentDeferred = defer.Deferred()
+        clientFactory = smtp.SMTPSenderFactory(
+            "bob@example.org", recipient + "@example.com",
+            StringIO(message), sentDeferred)
+        clientFactory.domain = "example.org"
+        clientConnector = reactor.connectTCP(
+            serverHost.host, serverHost.port, clientFactory)
+        self.addCleanup(clientConnector.disconnect)
+
+        def cbSent(ignored):
+            """
+            Verify that the message was successfully delivered and flush the
+            error which caused the first attempt to fail.
+            """
+            self.assertEquals(
+                domain.messages,
+                {recipient: ["\n%s\n" % (message,)]})
+            # Flush the RuntimeError that BrokenMessage caused to be logged.
+            self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 1)
+        sentDeferred.addCallback(cbSent)
+        return sentDeferred
+
+
 
 class SingletonRealm(object):
     """
@@ -948,6 +1354,7 @@ class ESMTPAuthenticationTestCase(unittest.TestCase):
         self.assertServerResponse('\r\n', [])
         self.assertServerAuthenticated(loginArgs, password='')
 
+
     def test_plainAuthenticationInitialResponse(self):
         """
         The response to the first challenge may be included on the AUTH command
@@ -1017,3 +1424,102 @@ class ESMTPAuthenticationTestCase(unittest.TestCase):
             ['501 Syntax error in parameters or arguments'])
 
         self.assertEqual(loginArgs, [])
+
+
+    def test_unexpectedLoginFailure(self):
+        """
+        If the L{Deferred} returned by L{Portal.login} fires with an
+        exception of any type other than L{UnauthorizedLogin}, the exception
+        is logged and the client is informed that the authentication attempt
+        has failed.
+        """
+        loginArgs = []
+        self.server.portal = self.portalFactory(loginArgs)
+
+        self.server.dataReceived('EHLO\r\n')
+        self.transport.clear()
+
+        self.assertServerResponse(
+            'AUTH LOGIN ' + 'username'.encode('base64').strip() + '\r\n',
+            ['334 ' + 'Password\0'.encode('base64').strip()])
+        self.assertServerResponse(
+            'password'.encode('base64').strip() + '\r\n',
+            [])
+
+        d, credentials, mind, interfaces = loginArgs.pop()
+        d.errback(RuntimeError("Something wrong with the server"))
+
+        self.assertEquals(
+            '451 Requested action aborted: local error in processing\r\n',
+            self.transport.value())
+
+        self.assertEquals(len(self.flushLoggedErrors(RuntimeError)), 1)
+
+
+
+class SMTPClientErrorTestCase(unittest.TestCase):
+    """
+    Tests for L{smtp.SMTPClientError}.
+    """
+    def test_str(self):
+        """
+        The string representation of a L{SMTPClientError} instance includes
+        the response code and response string.
+        """
+        err = smtp.SMTPClientError(123, "some text")
+        self.assertEquals(str(err), "123 some text")
+
+
+    def test_strWithNegativeCode(self):
+        """
+        If the response code supplied to L{SMTPClientError} is negative, it
+        is excluded from the string representation.
+        """
+        err = smtp.SMTPClientError(-1, "foo bar")
+        self.assertEquals(str(err), "foo bar")
+
+
+    def test_strWithLog(self):
+        """
+        If a line log is supplied to L{SMTPClientError}, its contents are
+        included in the string representation of the exception instance.
+        """
+        log = LineLog(10)
+        log.append("testlog")
+        log.append("secondline")
+        err = smtp.SMTPClientError(100, "test error", log=log.str())
+        self.assertEquals(
+            str(err),
+            "100 test error\n"
+            "testlog\n"
+            "secondline\n")
+
+
+
+class SenderMixinSentMailTests(unittest.TestCase):
+    """
+    Tests for L{smtp.SenderMixin.sentMail}, used in particular by
+    L{smtp.SMTPSenderFactory} and L{smtp.ESMTPSenderFactory}.
+    """
+    def test_onlyLogFailedAddresses(self):
+        """
+        L{smtp.SenderMixin.sentMail} adds only the addresses with failing
+        SMTP response codes to the log passed to the factory's errback.
+        """
+        onDone = self.assertFailure(defer.Deferred(), smtp.SMTPDeliveryError)
+        onDone.addCallback(lambda e: self.assertEquals(
+                e.log, "bob@example.com: 199 Error in sending.\n"))
+
+        clientFactory = smtp.SMTPSenderFactory(
+            'source@address', 'recipient@address',
+            StringIO("Message body"), onDone,
+            retries=0, timeout=0.5)
+
+        client = clientFactory.buildProtocol(
+            address.IPv4Address('TCP', 'example.net', 25))
+
+        addresses = [("alice@example.com", 200, "No errors here!"),
+                     ("bob@example.com", 199, "Error in sending.")]
+        client.sentMail(199, "Test response", 1, addresses, client.log)
+
+        return onDone
